from pathlib import Path
import pickle
from flax.training.train_state import TrainState
from typing import Literal

from mynx.callbacks import Callback
from mynx.logs import Logs

class CheckPoint(Callback):
    def __init__(self, path:Path, run_on:Literal["epoch", "step"]="epoch", freq:int=1):
        self.path = path
        self.run_on = run_on
        self.freq = freq
        super().__init__()

    def on_epoch_end(self, epoch:int, logs:Logs):
        if self.run_on == "step" or self.run_on == "epoch" and epoch % self.freq == 0:
            self.save(logs)

    def on_step_end(self, step:int, logs:Logs):
        if self.run_on == "step" and step % self.freq == 0:
            self.save(logs)

    def save(self, logs:Logs):
        if not self.path.parent.exists():
            self.path.parent.mkdir()
        with open(self.path, "wb") as f:
            pickle.dump(
                (
                    logs.state.step,
                    logs.state.params,
                    logs.state.opt_state
                ), 
                f
            )

    def on_train_start(self, logs:Logs):
        if not self.path.exists():
            return
        with open(self.path, "rb") as f:
            step, params, opt_state = pickle.load(f)
        logs.state = TrainState(step=step, apply_fn=logs.state.apply_fn, params=params, tx=logs.state.tx, opt_state=opt_state)


class Cycle:
    def __init__(self, iterable):
        self.data = [i for i in iterable]
        self.idx = 0

    def __next__(self):
        out = self.data[self.idx]
        self.idx += 1
        if self.idx >= len(self.data):
            self.idx = 0
        return out


from flax import linen as nn
from jax import numpy as jnp
from flax.typing import Array, Initializer
from typing import Optional
import einops as eo

class Attn(nn.Module):

    @nn.compact
    def __call__(self, q: Array, k: Array, v: Array, mask: Optional[Array] = None) -> Array:
        dims = q.shape[-1]

        scores = jnp.einsum("... j k, ... i k -> ... j i", q, k)
        scores /= dims
        if mask is not None:
            scores = jnp.where(mask, scores, -jnp.inf)
        scores = nn.softmax(scores)
        out = jnp.einsum("... j k, ... k i -> ... j i", scores, v)
        return out

class GroupedQuaryAttention(nn.Module):
    dim: int
    num_qheads: int
    num_kvheads: int
    use_bias: bool = True
    kernel_init: Initializer = nn.initializers.lecun_normal()
    bias_init: Initializer = nn.initializers.zeros_init()

    @nn.compact
    def __call__(self, q: Array, k: Optional[Array] = None, v: Optional[Array] = None, mask: Optional[Array] = None) -> Array:
        assert self.num_qheads % self.num_kvheads == 0, "num of kv heads must by divisible by num of q heads"
        assert self.dim % self.num_qheads == 0, "dims must be devisible by num of q heads"

        in_dim = q.shape[-1]

        head_dim = self.dim // self.num_qheads

        def linear(x, heads):
            x = nn.Dense(heads * head_dim, use_bias=self.use_bias, kernel_init=self.kernel_init, bias_init=self.bias_init)(x)
            x = eo.rearrange(x, "... (i j) -> ... i j", i=heads)
            return x

        if not k:
            k = q
        if not v:
            v = k

        qx = linear(q, self.num_qheads)
        kx = linear(k, self.num_kvheads)
        vx = linear(v, self.num_kvheads)

        qx = eo.rearrange(vx, "... i j k -> ... j i k")
        kx = eo.repeat(kx, "... i j k -> ... (j r) i k", r=self.num_qheads // self.num_kvheads)
        vx = eo.repeat(vx, "... i j k -> ... (j r) i k", r=self.num_qheads // self.num_kvheads)

        out = Attn()(qx, kx, vx, mask)

        out = eo.rearrange(out, "... i j k -> ... j (i k)")
        out = nn.Dense(in_dim, use_bias=self.use_bias, kernel_init=self.kernel_init, bias_init=self.bias_init)(out)

        return out

class MultiHeadAttention(nn.Module):
    dim: int
    num_heads: int
    use_bias: bool = True
    kernel_init: Initializer = nn.initializers.lecun_normal()
    bias_init: Initializer = nn.initializers.zeros_init()

    @nn.compact
    def __call__(self, q: Array, k: Optional[Array] = None, v: Optional[Array] = None, mask: Optional[Array] = None) -> Array:
        return GroupedQuaryAttention(self.dim, self.num_heads, self.num_heads, self.use_bias, self.kernel_init, self.bias_init)(q, k, v, mask)

class Attention(nn.Module):
    dim: int
    use_bias: bool = True
    kernel_init: Initializer = nn.initializers.lecun_normal()
    bias_init: Initializer = nn.initializers.zeros_init()

    @nn.compact
    def __call__(self, q: Array, k: Optional[Array] = None, v: Optional[Array] = None, mask: Optional[Array] = None) -> Array:
        return MultiHeadAttention(self.dim, 1, self.use_bias, self.kernel_init, self.bias_init)(q, k, v, mask)



from math import ceil

from mynx.callbacks import Callback
from mynx.logs import Logs

class Loading(Callback):
    def __init__(self, epoch_steps:int):
        self.epoch_steps = epoch_steps

    def on_step_end(self, batch:int, logs:Logs):
        batch = int((batch + 1) / self.epoch_steps * 20)
        loading = "=" * batch
        if (20 - batch) != 0:
            loading += ">"
        loading += " " * (20 - batch - 1)
        return f"[{loading}]"


from time import time
from datetime import timedelta

from mynx.callbacks import Callback
from mynx.logs import Logs

class TimeTracing(Callback):
    def __init__(self, epoch_steps:int):
        self.start_time = 0
        self.epoch_steps = epoch_steps

    def get_highst_unit(self, t:timedelta):
        sec = t.total_seconds()
        if sec < 1 / 1000:
            return f"{round(sec * 1000 ** 2)} us"
        if sec < 1:
            return f"{round(sec * 1000)} ms"
        if sec < 60:
            return f"{round(sec)} s"
        if sec < 3600:
            return f"{round(sec // 60)} m"
        if sec < 86400:
            return f"{round(sec // 3600)} h"
        return f"{t.days} d"

    def on_epoch_start(self, epoch:int, logs:Logs):
        self.start_time = time()

    def on_step_end(self, batch:int, logs:Logs):
        time_to_step = timedelta(seconds=(time() - self.start_time) / (batch + 1))
        time_to_end = time_to_step * (self.epoch_steps - batch - 1)
        return f"ETA: {self.get_highst_unit(time_to_end)} - {self.get_highst_unit(time_to_step)}/step"


from mynx.logs import Logs

class Callback:
    def __init__(self):
        ...

    def on_train_start(self, logs:Logs) -> None | str:
        ...

    def on_train_end(self, logs:Logs):
        ...

    def on_step_start(self, step:int, logs:Logs):
        ...

    def on_step_end(self, step:int, logs:Logs):
        ...

    def on_epoch_start(self, epoch:int, logs:Logs) -> None | str:
        ...

    def on_epoch_end(self, epoch:int, logs:Logs):
        ...


from mynx.callbacks import Callback
from mynx.logs import Logs

class StepCounter(Callback):
    def __init__(self, epoch_steps:int):
        self.epoch_steps = epoch_steps

    def on_step_end(self, batch:int, logs:Logs):
        return f"{batch + 1}/{self.epoch_steps}"


from abc import ABC, abstractmethod
import multiprocessing as mp
from time import sleep
import warnings

from mynx.callbacks import Callback
from mynx.logs import Logs
from mynx.src import Cycle

class DataLoader(Callback, ABC):
    def __init__(self, use_multiprocesing:bool = True, num_workers:int = None, max_queued_batches: int = 8, warmup_queue:bool = True, disable_warnings:bool = False):
        if not num_workers:
            num_workers = mp.cpu_count()
        self.use_multiprocesing = use_multiprocesing
        self.num_workers = num_workers
        self.max_queued_batches = max_queued_batches
        self.warmup_queue = warmup_queue
        self.disable_warnings = disable_warnings

        self.get_batch_idx = Cycle(range(len(self)))

        if use_multiprocesing:
            self._batch_queue = mp.Queue(self.max_queued_batches)
            self._task_queues = [mp.Queue() for _ in range(num_workers)]

            self._workers = [mp.Process(target=self._worker, args=(queue,), daemon=True) for queue in self._task_queues]

            for worker in self._workers:
                worker.start()

    def start(self):
        if not self.use_multiprocesing:
            return
        for _ in range(self.max_queued_batches):
            batch_idx = next(self.get_batch_idx)
            self._task_queues[batch_idx % self.num_workers].put(batch_idx)

        while self.warmup_queue and not self._batch_queue.full():
            sleep(0.1)

    @abstractmethod
    def __len__(self):
        ...

    @abstractmethod
    def get_batch(self, idx:int):
        ...

    def _worker(self, queue:mp.Queue):
        while True:
            batch_idx = queue.get()
            batch = self.get_batch(batch_idx)
            self._batch_queue.put(batch)

    def on_step_start(self, step:int, logs:Logs):
        logs.batch = next(self)

    def __next__(self):
        if not self.use_multiprocesing:
            return self.get_batch(next(self.get_batch_idx))
        
        if not self.disable_warnings and self._batch_queue.qsize() == 0:
            warnings.warn(f"Batches are not preparing fast enought. Consider optimizing `{self.__class__.__name__}.{self.get_batch.__name__}` method")
        
        batch = self._batch_queue.get()
        batch_idx = next(self.get_batch_idx)
        self._task_queues[batch_idx % self.num_workers].put(batch_idx)
        return batch
    
    def terminate(self):
        if not self.use_multiprocesing:
            return
        
        for worker in self._workers:
            worker.terminate()
            worker.join()


from mynx.callbacks import Callback
from mynx.logs import Logs

class EpochCounter(Callback):
    def __init__(self, epochs:int):
        self.epochs = epochs

    def on_epoch_start(self, epoch:int, logs:Logs):
        return f"Epoch: {epoch + 1}/{self.epochs}"


from flax import linen as nn
from flax.training.train_state import TrainState
from flax.training import common_utils
from flax import jax_utils
import optax
import jax.numpy as jnp
import jax
import math
from typing import Callable

from mynx import Logs, DataLoader
from mynx.callbacks import Callback
from mynx.callbacks.metrics import EpochCounter, StepCounter, Loading, TimeTracing, Loss

class Model:
    def __init__(self, nn:nn.Module, loss:Callable, data_loader:DataLoader) -> None:
        self.nn = nn
        self.loss = loss
        self.data_loader = data_loader
        self.params = None

    def _step_fn(self, params, state, x, y_true):
        y_pred = state.apply_fn({"params":params}, x)
        loss = self.loss(y_pred, y_true)
        return loss, y_pred

    def _tran_step(self, state, x, y_true):
        grad_fn = jax.value_and_grad(self._step_fn, has_aux=True, allow_int=True)
        (loss, y_pred), grads = grad_fn(state.params, state, x, y_true)
        grads = jax.lax.pmean(grads, axis_name="devices")
        state = state.apply_gradients(grads=grads)
        return loss, y_pred, state

    def tabulate(self, depth:int=1):
        print(nn.tabulate(self.nn, jax.random.PRNGKey(0), depth=depth)(self.data_loader.get_batch(0)[0]))

    def get_params(self):
        if not self.params:
            self.params = self.nn.init(jax.random.PRNGKey(0), self.data_loader.get_batch(0)[0])["params"]
        return self.params
    
    def _fit_on_train_start(self, tx, epochs, total_steps, callbecks, default_callbecks):
        epoch_steps = len(self.data_loader)
        if total_steps:
            epochs = math.ceil(total_steps / epoch_steps)
        elif not epochs:
            raise ValueError("Missing argument epochs or total_steps")
        total_steps = epochs * epoch_steps
        
        self.callbecks:list[Callback] = [self.data_loader]
        if default_callbecks:
            self.callbecks += [
                EpochCounter(epochs),
                StepCounter(epoch_steps),
                Loading(epoch_steps),
                TimeTracing(epoch_steps),
                Loss()
            ]
        self.callbecks += callbecks

        self.state = TrainState.create(apply_fn=self.nn.apply, params=self.get_params(), tx=tx)

        self._tran_step = jax.pmap(self._tran_step, axis_name="devices")

        logs = Logs(state=self.state)

        for callbeck in self.callbecks:
            if callbeck_msg := callbeck.on_train_start(logs):
                print(callbeck_msg)

        total_start_step = logs.state.step
        start_epoch = total_start_step // epoch_steps
        start_step = total_start_step % epoch_steps
        self.data_loader.get_batch_idx.idx = start_step
        self.data_loader.start()

        self.device_state = jax_utils.replicate(logs.state)

        return epochs, logs, start_epoch, start_step

    def _fit_on_epoch_start(self, epoch, logs):
        self.msg = []
        for callbeck in self.callbecks:
            if callbeck_msg := callbeck.on_epoch_start(epoch, logs):
                self.msg.append(callbeck_msg)
        if self.msg != []:
            self.msg = " - ".join(self.msg)
            print(self.msg)

    def _fit_on_step_start(self, step, logs):
        for callbeck in self.callbecks:
            callbeck.on_step_start(step, logs)
        return logs.batch

    def _fit_step(self, batch, logs):
        shard_batch = jax.tree_util.tree_map(common_utils.shard, batch)
        loss, y_pred, self.device_state = self._tran_step(self.device_state, *shard_batch)
        loss = jnp.mean(loss)
        logs.loss = loss
        logs.y_pred = jnp.concatenate(y_pred, axis=1)
        logs.state = jax_utils.unreplicate(self.device_state)

    def _fit_on_step_end(self, step, logs):
        last_msg_len = len(self.msg)
        self.msg = []
        for callbeck in self.callbecks:
            if callbeck_msg := callbeck.on_step_end(step, logs):
                self.msg.append(callbeck_msg)
        if self.msg != []:
            self.msg = " - ".join(self.msg)
            if len(self.msg) < last_msg_len:
                self.msg += " " * (last_msg_len - len(self.msg))
            print("\r" + self.msg, end="")

    def _fit_on_epoch_end(self, epoch, logs):
        print()
        self.msg = []
        for callbeck in self.callbecks:
            if callbeck_msg := callbeck.on_epoch_end(epoch, logs):
                self.msg.append(callbeck_msg)
        if self.msg != []:
            self.msg = " - ".join(self.msg)
            print(self.msg)
    
    def _fit_on_train_end(self, logs):
        for callbeck in self.callbecks:
            if callbeck_msg := callbeck.on_train_end(logs):
                print(callbeck_msg)

    def fit(self, tx:optax.GradientTransformation, epochs:int=None, total_steps:int=None, callbecks:list[Callback]=[], default_callbecks:bool=True):
        epochs, logs, start_epoch, start_step = self._fit_on_train_start(tx, epochs, total_steps, callbecks, default_callbecks)

        for epoch in range(start_epoch, epochs):
            self._fit_on_epoch_start(epoch, logs)

            for step in range(start_step, len(self.data_loader)):
                batch = self._fit_on_step_start(step, logs)

                self._fit_step(batch, logs)

                self._fit_on_step_end(step, logs)
            
            self._fit_on_epoch_end(epoch, logs)
            
            start_step = 0

        self._fit_on_train_end(logs)
        
        self.state = logs.state


